from __future__ import annotations

from contextlib import contextmanager
from typing import (
    cast,
    Dict,
    Mapping,
    Optional,
    TYPE_CHECKING,
)


if TYPE_CHECKING:  # pragma: no cover
    from amalgam.amalgams import Amalgam

    Bindings = Mapping[str, Amalgam]


class SymbolNotFound(Exception):
    """Synonym for :class:`KeyError`."""


class TopLevelPop(Exception):
    """Raised at :meth:`Environment.env_pop`."""


class Environment:
    """
    Class that manages and represents nested execution environments.

    Attributes:
      bindings (:class:`Dict[str, Amalgam]`): A mapping of
        :class:`str` keys to :class:`.amalgams.Amalgam` values.

      parent (:class:`Optional[Environment]`): The parent
        :class:`Environment` instance to search into, forming a
        linked list.

      level (:class:`int`): The current length of the
        :class:`Environment` linked list. If a
        :attr:`~.Environment.parent` is provided, sets the current
        value to the parent's :attr:`~.Environment.level` + 1.

      search_depth (:class:`int`): The search depth when traversing
        the :class:`Environment` linked list in the
        :meth:`~.Environment.__contains__`,
        :meth:`~.Environment.__delitem__`,
        :meth:`~.Environment.__getitem__`, and
        :meth:`~.Environment.__setitem__` methods.
    """

    def __init__(
        self,
        bindings: Bindings = None,
        parent: Environment = None,
    ) -> None:
        self.bindings: Dict[str, Amalgam] = {**bindings} if bindings else {}
        self.parent: Optional[Environment] = parent
        self.level: int = parent.level + 1 if parent else 0
        self.search_depth: int = 0

    def __getitem__(self, item: str) -> Amalgam:
        """
        Attempts to recursively obtain the provided `item`.

        Searches with respect to the current :attr:`search_depth` of the
        calling :class:`Environment` instance. If an existing `item`
        is encountered at a certain depth less than the target depth,
        returns that `item`, otherwise, raises :class:`SymbolNotFound`.
        """
        _self = self

        if self.search_depth >= 0:
            depth = self.search_depth + 1
        else:
            depth = self.level + 1

        # Search until the top-most environment,
        # raise SymbolNotFound otherwise.
        for _ in range(depth):
            try:
                return _self.bindings[item]
            except KeyError:
                _self = cast(Environment, _self.parent)
        else:
            raise SymbolNotFound(item)

    def __setitem__(self, item: str, value: Amalgam) -> None:
        """
        Attempts to recursively set the provided `value` to an `item`.

        Searches with respect to the current :attr:`search_depth` of the
        calling :class:`Environment` instance. If an existing `item` is
        encountered at a certain depth less than the target depth,
        overrides that `item` instead.
        """
        _self = self

        if self.search_depth >= 0:
            depth = self.search_depth
        else:
            depth = self.level

        # Search until the second-to-last environment,
        # set at the top-most environment otherwise.
        for _ in range(depth):
            if item in _self.bindings:
                _self.bindings[item] = value
                break
            else:
                _self = cast(Environment, _self.parent)
        else:
            _self.bindings[item] = value

    def __delitem__(self, item: str) -> None:
        """
        Attempts to recursively delete the provided `item`.

        Searches with respect to the current :attr:`search_depth` of the
        calling :class:`Environment` instance. If an existing `item` is
        encountered at a certain depth less than the target depth,
        deletes that `item` instead.
        """
        _self = self

        if self.search_depth >= 0:
            depth = self.search_depth + 1
        else:
            depth = self.level + 1

        # Search until the top-most environment,
        # raise SymbolNotFound otherwise.
        for _ in range(depth):
            try:
                del _self.bindings[item]
                break
            except KeyError:
                _self = cast(Environment, _self.parent)
        else:
            raise SymbolNotFound(item)

    def __contains__(self, item: str) -> bool:
        """
        Recursively checks whether an `item` exists.

        Searches with respect to the current :attr:`search_depth` of the
        calling :class:`Environment` instance. If the target `item` is
        encountered at a certain depth less than the target depth,
        immediately returns `True`, otherwise, returns `False`.
        """
        _self = self

        if self.search_depth >= 0:
            depth = self.search_depth + 1
        else:
            depth = self.level + 1

        # Search until an `item` is found,
        # return False otherwise
        for _ in range(depth):
            if item in _self.bindings:
                return True
            else:
                _self = cast(Environment, _self.parent)
        else:
            return False

    @contextmanager
    def search_at(self, *, depth=0):
        """
        Context manager for temporarily setting the lookup depth.

        The provided `depth` argument must not exceed the :attr:`level`
        of the calling :class:`Environment` instance, and will raise a
        :class:`ValueError` if done so.

        >>> env = Environment(FUNCTIONS)
        >>>
        >>> with env.search_at(depth=42):
        ...     env["+"]  # Raises ValueError

        Any negative integer can be passed as a `depth` to signify
        an infinite lookup until the top-most environment.

        >>> env = Environment(FUNCTIONS)
        >>> cl_env = env.env_push({...})
        >>>
        >>> with cl_env.search_at(depth=-1):
        ...    cl_env["+"]  # Searches `env`
        """
        if depth > self.level:
            exc = ValueError(
                f"depth {depth} is greater than maximum level {self.level}"
            )
            raise exc

        self.search_depth = depth

        try:
            yield self
        finally:
            self.search_depth = 0

    def env_push(self, bindings: Bindings = None) -> Environment:
        """
        Creates a new :class:`Environment` and binds the calling
        instance as its parent environment.
        """
        return Environment(bindings, self)

    def env_pop(self) -> Environment:
        """
        Discards the current :class:`Environment` and returns the parent
        :class:`Environment`.
        """
        if self.parent is not None:
            return self.parent
        else:
            raise TopLevelPop("cannot discard top-level Environment")



import click

import amalgam.engine as en


@click.command()
@click.argument("file", required=False, type=click.File("r"))
@click.option("-e", "--expr", default=None, help="Expression to evaluate.")
def amalgam_main(file, expr):
    has_file = file is not None
    has_expr = expr is not None

    if not has_file and not has_expr:
        en.Engine().repl()  # Guarantees SystemExit

    elif has_file and not has_expr:
        text = file.read()

    elif not has_file and has_expr:
        text = expr

    else:
        raise click.BadParameter("Cannot use FILE and --expr together")

    result = en.Engine().parse_and_run(text)

    if not has_file and has_expr:
        print(result)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from fractions import Fraction
from typing import (
    cast,
    Any,
    Callable,
    Generic,
    Mapping,
    Sequence,
    Tuple,
    TypeVar,
)

import amalgam.environment as ev


class Amalgam(ABC):
    """The abstract base class for language constructs."""

    @abstractmethod
    def evaluate(self, environment: ev.Environment) -> Any:
        """
        Protocol for evaluating or unwrapping :class:`Amalgam` objects.
        """

    def bind(self, environment: ev.Environment) -> Amalgam:  # pragma: no cover
        """
        Protocol for implementing environment binding for
        :class:`Function`.

        This base implementation is responsible for allowing `bind`
        to be called on other :class:`Amalgam` subclasses by
        performing no operation aside from returning :data:`self`.
        """
        return self

    def call(
        self, environment: ev.Environment, *arguments: Amalgam
    ) -> Amalgam:  # pragma: no cover
        """
        Protocol for implementing function calls for
        :class:`Function`.

        This base implementation is responsible for making the type
        signature of :attr:`SExpression.func` to properly type check
        when :meth:`SExpression.evaluate` is called, as well as raising
        :class:`NotImplementedError` for non-callable types.
        """
        raise NotImplementedError(f"{self.__class__.__name__} is not callable")

    def _make_repr(self, value: Any) -> str:  # pragma: no cover
        """Helper method for creating a :meth:`__repr__`."""
        return f"<{self.__class__.__name__} '{value!s}' @ {hex(id(self))}>"


@dataclass(repr=False)
class Atom(Amalgam):
    """
    An :class:`.Amalgam` that represents different atoms.

    Attributes:
      value (:class:`str`): The name of the atom.
    """

    value: str

    def evaluate(self, _environment: ev.Environment) -> Atom:
        """Evaluates to the same :class:`.Atom` reference."""
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(self.value)

    def __str__(self) -> str:
        return f":{self.value}"


N = TypeVar("N", int, float, Fraction)


@dataclass(repr=False, order=True)
class Numeric(Amalgam, Generic[N]):
    """
    An :class:`.Amalgam` that wraps around numeric types.

    Parameterized as a :class:`Generic` by:
    :data:`N = TypeVar("N", int, float, Fraction)`

    Attributes:
      value (:data:`N`): The numeric value being wrapped.
    """

    value: N

    def evaluate(self, _environment: ev.Environment) -> Numeric:
        """Evaluates to the same :class:`.Numeric` reference."""
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(self.value)

    def __str__(self) -> str:
        return str(self.value)


@dataclass(repr=False, order=True)
class String(Amalgam):
    """
    An :class:`.Amalgam` that wraps around strings.

    Attributes:
      value (:class:`str`): The string being wrapped.
    """

    value: str

    def evaluate(self, _environment: ev.Environment) -> String:
        """Evaluates to the same :class:`.String` reference."""
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(f"\"{self.value}\"")

    def __str__(self) -> str:
        return f"\"{self.value}\""


@dataclass(repr=False)
class Symbol(Amalgam):
    """
    An :class:`.Amalgam` that wraps around symbols.

    Attributes:
      value (:class:`str`): The name of the symbol.
    """

    value: str

    def evaluate(self, environment: ev.Environment) -> Amalgam:
        """
        Searches the provided `environment` fully with
        :attr:`Symbol.value`. Returns the :class:`.Amalgam` object
        bound to the :attr:`Symbol.value` in the environment. Raises
        :class:`.environment.SymbolNotFound` if a binding is not found.
        """
        with environment.search_at(depth=-1):
            return environment[self.value]

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(self.value)

    def __str__(self) -> str:
        return self.value


class DisallowedContextError(Exception):
    """Raised on functions outside of their intended contexts."""


@dataclass(repr=False)
class Function(Amalgam):
    """
    An :class:`.Amalgam` that wraps around functions.

    Attributes:
      name (:class:`str`): The name of the function.

      fn (:class:`Callable[..., Amalgam]`): The function being wrapped.
        Must have the signature: `(env, amalgams...) -> amalgam`.

      defer (:class:`bool`): If set to :obj:`True`, arguments are
        wrapped in :class:`.Quoted` before being passed to
        :attr:`.Function.fn`.

      contextual (:class:`bool`): If set to :obj:`True`, disallows
        function calls when :attr:`.Function.in_context` is set to
        :obj:`False`.

      env (:class:`.environment.Environment`): The
        :class:`.environment.Environment` instance bound to the
        function. Overrides the `environment` parameter passed to the
        :meth:`.Function.call` method.

      in_context (:class:`bool`): Predicate that disallows functions
        to be called outside of specific contexts. Makes
        :meth:`.Function.call` raise :class:`.DisallowedContextError`
        when set to :obj:`False` and :attr:`.Function.contextual` is
        set to :obj:`True`.
    """

    name: str
    fn: Callable[..., Amalgam]
    defer: bool = False
    contextual: bool = False

    def __post_init__(self):
        self.env = cast(ev.Environment, None)
        self.in_context = False

    def evaluate(self, _environment: ev.Environment) -> Function:
        """Evaluates to the same :class:`.Function` reference."""
        return self

    def bind(self, environment: ev.Environment) -> Function:
        """
        Sets the :attr:`.Function.env` attribute and returns the same
        :class:`.Function` reference.
        """
        self.env = environment
        return self

    def call(self, environment: ev.Environment, *arguments: Amalgam) -> Amalgam:
        """
        Performs the call to the :attr:`.Function.fn` attribute.

        Performs pre-processing depending on the values of
        :attr:`.Function.defer`, :attr:`.Function.contextual`, and
        :attr:`.Function.in_context`,
        """
        if self.contextual and not self.in_context:
            raise DisallowedContextError(f"invalid context for {self.name}")

        if self.env is not None:
            environment = self.env

        if self.defer:
            arguments = tuple(Quoted(arg) for arg in arguments)
        else:
            arguments = tuple(arg.evaluate(environment) for arg in arguments)

        return self.fn(environment, *arguments)

    def with_name(self, name: str) -> Function:
        """
        Sets the :attr:`.Function.name` attribute and returns the same
        :class:`.Function` reference.
        """
        self.name = name
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(self.name)

    def __str__(self) -> str:  # pragma: no cover
        return self.name


@dataclass(init=False, repr=False)
class SExpression(Amalgam):
    """
    An :class:`.Amalgam` that wraps around S-Expressions.

    Attributes:
      vals (:class:`Tuple[Amalgam, ...]`): Entities contained by the
        S-Expression.
    """

    vals: Tuple[Amalgam, ...]

    def __init__(self, *vals: Amalgam) -> None:
        self.vals = vals

    @property
    def func(self) -> Amalgam:
        """The head of the :attr:`SExpression.vals`."""
        return self.vals[0]

    @property
    def args(self) -> Tuple[Amalgam, ...]:
        """The rest of the :attr:`SExpression.vals`."""
        return self.vals[1:]

    def evaluate(self, environment: ev.Environment) -> Amalgam:
        """
        Evaluates :attr:`func` using `environment` before invoking
        the :meth:`call` method with `environment` and
        :attr:`SExpression.args`.
        """
        return self.func.evaluate(environment).call(environment, *self.args)

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(f"{self.func!r} {' '.join(map(repr, self.args))}")

    def __str__(self) -> str:
        return f"({' '.join(map(str, self.vals))})"


T = TypeVar("T", bound=Amalgam)


@dataclass(init=False, repr=False)
class Vector(Amalgam, Generic[T]):
    """
    An :class:`.Amalgam` that wraps around a homogenous vector.

    Parameterized as a :class:`Generic` by:
    :data:`T = TypeVar("T", bound=Amalgam)`

    Attributes:
      vals (:class:`Tuple[T, ...]`): Entities contained by the vector

      mapping (:class:`Mapping[str, Amalgam]`): Mapping representing
        vectors with :class:`.Atom` s for odd indices and
        :class:`.Amalgam` s for even indices.
    """

    vals: Tuple[T, ...]

    def __init__(self, *vals: T) -> None:
        self.vals = vals
        self.mapping = self._as_mapping()

    def evaluate(self, environment: ev.Environment) -> Vector:
        """
        Creates a new :class:`.Vector` by evaluating every value in
        :attr:`Vector.vals`.
        """
        return Vector(*(val.evaluate(environment) for val in self.vals))

    def _as_mapping(self) -> Mapping[str, Amalgam]:
        """
        Attemps to create a :class:`Mapping[str, Amalgam]` from
        :attr:`Vector.vals`.

        Odd indices must be :class:`.Atom` s and even indices must be
        :class:`.Amalgam` s. Returns an empty mapping if this form is
        not met.
        """
        if len(self.vals) % 2 != 0 or len(self.vals) == 0:
            return {}

        mapping = {}

        atoms = self.vals[::2]
        amalgams = self.vals[1::2]

        for atom, amalgam in zip(atoms, amalgams):
            if not isinstance(atom, Atom):
                return {}
            mapping[atom.value] = amalgam

        return mapping

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(" ".join(map(repr, self.vals)))

    def __str__(self) -> str:
        return f"[{' '.join(map(str, self.vals))}]"


@dataclass(repr=False)
class Quoted(Amalgam, Generic[T]):
    """
    An :class:`Amalgam` that defers evaluation of other
    :class:`Amalgam` s.

    Parameterized as a :class:`Generic` by:
    :data:`T = TypeVar("T", bound=Amalgam)`

    Attributes:
      value (:data:`T`): The :class:`.Amalgam` being deferred.
    """

    value: T

    def evaluate(self, _environment: ev.Environment) -> Quoted:
        """Evaluates to the same :class:`.Quoted` reference."""
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(repr(self.value))

    def __str__(self) -> str:
        return f"'{self.value!s}"


P = TypeVar("P", bound=object)


@dataclass(repr=False)
class Internal(Amalgam, Generic[P]):
    """
    An :class:`Amalgam` that holds Python :class:`object` s.

    Parameterized as a :class:`Generic` by:
    :data:`P = TypeVar("P", bound=object)`

    Attributes:
      value (:data:`P`): The Python :class:`object` being wrapped.
    """

    value: P

    def evaluate(self, _environment: ev.Environment) -> Internal:
        """Evaluates to the same :class:`.Internal` reference."""
        return self

    def __repr__(self) -> str:  # pragma: no cover
        return self._make_repr(repr(self.value))

    def __str__(self) -> str:  # pragma: no cover
        return f"~{self.value!s}~"


def create_fn(
    fname: str,
    fargs: Sequence[str],
    fbody: Amalgam,
    defer: bool = False,
    contextual: bool = False,
) -> Function:
    """Helper function for creating `Function` objects.

    Given the name of the function: `fname`, a sequence of argument
    names: `fargs`, and the `Amalgam` to be evaluated: `fbody`,
    creates a new `closure_fn` to be wrapped by a `Function`.
    """

    def closure_fn(environment: ev.Environment, *arguments: Amalgam) -> Amalgam:
        """Callable responsible for evaluating `fbody`."""

        # Create a child environment and bind args to their names.
        # TODO: Raise an error when missing arguments instead.
        cl_env = environment.env_push(dict(zip(fargs, arguments)))

        # Call the `evaluate` method on the function body with
        # `cl_env` and then call `bind` on the result with the
        # same environment.
        return fbody.evaluate(cl_env).bind(cl_env)

    return Function(fname, closure_fn, defer, contextual)



from fractions import Fraction
from functools import partial, wraps
from itertools import chain
from pathlib import Path
import sys
from typing import (
    cast, Callable, Dict, List, NamedTuple, Sequence, TypeVar, Union
)

import amalgam.amalgams as am
import amalgam.engine as en
import amalgam.environment as ev


FUNCTIONS: Dict[str, am.Function] = {}


T = TypeVar("T", bound=am.Amalgam)


def _make_function(
    name: str,
    func: Callable[..., T] = None,
    defer: bool = False,
    contextual: bool = False,
    allows: Sequence[str] = None,
) -> Union[partial, Callable[..., T]]:
    """
    Transforms a given function `func` into a `Function`
    and stores it inside of the `FUNCTIONS` mapping.
    """

    if func is None:
        return partial(
            _make_function,
            name,
            defer=defer,
            contextual=contextual,
            allows=allows,
        )

    if allows is None:
        allows = []

    @wraps(func)
    def _func(env, *arguments, **keywords):
        with env.search_at(depth=-1):
            fns = [env[allow] for allow in allows]

        for fn in fns:
            cast(am.Function, fn).in_context = True

        result = func(env, *arguments, **keywords)

        for fn in fns:
            cast(am.Function, fn).in_context = False

        return result

    FUNCTIONS[name] = am.Function(name, _func, defer, contextual)

    return _func


@_make_function("+")
def _add(_env: ev.Environment, *nums: am.Numeric) -> am.Numeric:
    """Returns the sum of :data:`nums`."""
    return am.Numeric(sum(num.value for num in nums))


@_make_function("-")
def _sub(_env: ev.Environment, *nums: am.Numeric) -> am.Numeric:
    """
    Subtracts :data:`nums[0]` and the summation of :data:`nums[1:]`.
    """
    x, *ns = (num.value for num in nums)
    y: Union[float, Fraction] = 0
    for n in ns:
        y += n
    return am.Numeric(x - y)


@_make_function("*")
def _mul(_env: ev.Environment, *nums: am.Numeric) -> am.Numeric:
    """Returns the product of :data:`nums`."""
    prod: Union[float, Fraction] = 1
    for num in nums:
        prod *= num.value
    return am.Numeric(prod)


@_make_function("/")
def _div(_env: ev.Environment, *nums: am.Numeric) -> am.Numeric:
    """
    Divides :data:`nums[0]` and the product of :data:`nums[1:]`
    """
    x, *ns = (num.value for num in nums)
    y: Union[float, Fraction] = 1
    for n in ns:
        y *= n
    return am.Numeric(x / y)


@_make_function("setn", defer=True)
def _setn(
    env: ev.Environment,
    name: am.Quoted[am.Symbol],
    amalgam: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Binds :data:`name` to the evaluated :data:`amalgam` value in the
    immediate :data:`env` and returns that value.
    """
    env[name.value.value] = amalgam.value.evaluate(env)
    return env[name.value.value]


@_make_function("fn", defer=True)
def _fn(
    env: ev.Environment,
    args: am.Quoted[am.Vector[am.Symbol]],
    body: am.Quoted[am.Amalgam],
) -> am.Function:
    """
    Creates an anonymous function using the provided arguments.

    Binds :data:`env` to the created :class:`.amalgams.Function` if a
    closure is formed.
    """
    fn = am.create_fn("~lambda~", [arg.value for arg in args.value.vals], body.value)
    if env.parent is not None:
        fn.bind(env)
    return fn


@_make_function("mkfn", defer=True)
def _mkfn(
    env: ev.Environment,
    name: am.Quoted[am.Symbol],
    args: am.Quoted[am.Vector[am.Symbol]],
    body: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Creates a named function using the provided arguments.

    Composes :func:`._fn` and :func:`._setn`.
    """
    return _setn(env, name, am.Quoted(_fn(env, args, body).with_name(name.value.value)))


@_make_function("let", defer=True)
def _let(
    env: ev.Environment,
    qpairs: am.Quoted[am.Vector[am.Vector]],
    body: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Creates temporary bindings of names to values specified in
    :data:`qpairs` before evaluating :data:`body`.
    """
    names = []
    values = []

    for pos, pair in enumerate(qpairs.value.vals):
        if not isinstance(pair, am.Vector) or len(pair.vals) != 2:
            raise ValueError(f"{pair} at {pos} is not a pair")

        name, value = pair.vals

        if not isinstance(name, am.Symbol):
            raise TypeError(f"{name} at {pos} is not a symbol")

        names.append(name)
        values.append(value)

    return _fn(env, am.Quoted(am.Vector(*names)), body).call(env, *values)


@_make_function("bool")
def _bool(_env: ev.Environment, expr: am.Amalgam) -> am.Atom:
    """Checks for the truthiness of an :data:`expr`."""
    if expr == am.String(""):
        return am.Atom("FALSE")
    elif expr == am.Numeric(0):
        return am.Atom("FALSE")
    elif expr == am.Vector():
        return am.Atom("FALSE")
    elif expr == am.Atom("FALSE"):
        return am.Atom("FALSE")
    elif expr == am.Atom("NIL"):
        return am.Atom("FALSE")
    return am.Atom("TRUE")


@_make_function(">")
def _gt(_env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs a `greater than` comparison."""
    if x > y:  # type: ignore
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("<")
def _lt(_env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs a `less than` comparison."""
    if x < y:  # type: ignore
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("=")
def _eq(_env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs an `equals` comparison."""
    if x == y:
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("/=")
def _ne(_env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs a `not equals` comparison."""
    if x != y:
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function(">=")
def _ge(env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs a `greater than or equal` comparison."""
    if x >= y:  # type: ignore
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("<=")
def _le(env: ev.Environment, x: am.Amalgam, y: am.Amalgam) -> am.Atom:
    """Performs a `less than or equal` comparison."""
    if x <= y:  # type: ignore
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("not")
def _not(_env: ev.Environment, expr: am.Amalgam) -> am.Atom:
    """Checks and negates the truthiness of :data:`expr`."""
    if _bool(_env, expr) == am.Atom("TRUE"):
        return am.Atom("FALSE")
    return am.Atom("TRUE")


@_make_function("and", defer=True)
def _and(env: ev.Environment, *qexprs: am.Quoted[am.Amalgam]) -> am.Atom:
    """
    Checks the truthiness of the evaluated :data:`qexprs` and performs
    an `and` operation. Short-circuits when :data:`:FALSE` is returned
    and does not evaluate subsequent expressions.
    """
    for qexpr in qexprs:
        cond = _bool(env, qexpr.value.evaluate(env))
        if cond == am.Atom("FALSE"):
            return cond
    return am.Atom("TRUE")


@_make_function("or", defer=True)
def _or(env: ev.Environment, *qexprs: am.Quoted[am.Amalgam]) -> am.Atom:
    """
    Checks the truthiness of the evaluated :data:`qexprs` and performs
    an `or` operation. Short-circuits when :data:`:TRUE` is returned
    and does not evaluate subsequent expressions.
    """
    for qexpr in qexprs:
        cond = _bool(env, qexpr.value.evaluate(env))
        if cond == am.Atom("TRUE"):
            return cond
    return am.Atom("FALSE")


@_make_function("if", defer=True)
def _if(
    env: ev.Environment,
    qcond: am.Quoted[am.Amalgam],
    qthen: am.Quoted[am.Amalgam],
    qelse: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Checks the truthiness of the evaluated :data:`qcond`, evaluates and
    returns :data:`qthen` if :data:`:TRUE`, otherwise, evaluates and
    returns :data:`qelse`.
    """
    cond = _bool(env, qcond.value.evaluate(env))
    if cond == am.Atom("TRUE"):
        return qthen.value.evaluate(env)
    return qelse.value.evaluate(env)


@_make_function("cond", defer=True)
def _cond(env: ev.Environment, *qpairs: am.Quoted[am.Vector[am.Amalgam]]) -> am.Amalgam:
    """
    Traverses pairs of conditions and values. If the condition evaluates
    to :data:`:TRUE`, returns the value pair and short-circuits
    evaluation. If no conditions are met, :data:`:NIL` is returned.
    """
    for qpair in qpairs:
        pred, expr = qpair.value.vals
        if _bool(env, pred.evaluate(env)) == am.Atom("TRUE"):
            return expr.evaluate(env)
    return am.Atom("NIL")


@_make_function("exit")
def _exit(env: ev.Environment, exit_code: am.Numeric = am.Numeric(0)) -> am.Amalgam:
    """Exits the program with the given :data:`exit_code`."""
    print("Goodbye.")
    sys.exit(int(exit_code.value))


@_make_function("print")
def _print(_env: ev.Environment, amalgam: am.Amalgam) -> am.Amalgam:
    """Prints the provided :data:`amalgam` and returns it."""
    print(amalgam)
    return amalgam


@_make_function("putstrln")
def _putstrln(_env: ev.Environment, string: am.String) -> am.String:
    """Prints the provided :data:`string` and returns it."""
    if not isinstance(string, am.String):
        raise TypeError("putstrln only accepts a string")
    print(string.value)
    return string


@_make_function("do", defer=True)
def _do(env: ev.Environment, *qexprs: am.Quoted[am.Amalgam]) -> am.Amalgam:
    """
    Evaluates a variadic amount of :data:`qexprs`, returning the final
    expression evaluated.
    """
    accumulator = am.Atom("NIL")
    for qexpr in qexprs:
        accumulator = qexpr.value.evaluate(env)
    return accumulator


@_make_function("require")
def _require(env: ev.Environment, module_name: am.String) -> am.Atom:
    """
    Runs a given :data:`module_name` and imports the exposed symbols to
    the current :data:`env` with respect to the `~provides~` key created
    in :func:`._provide`.
    """
    module_path = Path(module_name.value).absolute()
    with module_path.open("r", encoding="UTF-8") as f:
        text = f.read()

    snapshot = env.bindings.copy()

    internal_engine = cast(am.Internal[en.Engine], env["~engine~"])
    internal_engine.value.parse_and_run(text)

    if "~provides~" in env:
        symbols = cast(am.Vector[am.Symbol], env["~provides~"])
        exports = {symbol.value for symbol in symbols.vals}

        changes = {
            name: env[name]
            for name in exports.intersection(env.bindings)
        }

        snapshot.update(changes)
        env.bindings = snapshot

    return am.Atom("NIL")


@_make_function("provide", defer=True)
def _provide(env: ev.Environment, *qsymbols: am.Quoted[am.Symbol]) -> am.Atom:
    """Sets the `~provides~` key to be used in :func:`._require`."""
    env["~provides~"] = am.Vector(*(qsymbol.value for qsymbol in qsymbols))
    return am.Atom("NIL")


@_make_function("concat")
def _concat(_env: ev.Environment, *strings: am.String) -> am.String:
    """Concatenates the given :data:`strings`."""
    return am.String("".join(string.value for string in strings))


@_make_function("merge")
def _merge(_env: ev.Environment, *vectors: am.Vector) -> am.Vector:
    """Merges the given :data:`vectors`."""
    return am.Vector(*chain.from_iterable(vector.vals for vector in vectors))


@_make_function("slice")
def _slice(
    _env: ev.Environment,
    vector: am.Vector,
    start: am.Numeric,
    stop: am.Numeric,
    step: am.Numeric = am.Numeric(1),
) -> am.Vector:
    """Returns a slice of the given :data:`vector`."""
    return am.Vector(*vector.vals[start.value:stop.value:step.value])


@_make_function("at")
def _at(_env: ev.Environment, index: am.Numeric, vector: am.Vector) -> am.Amalgam:
    """Indexes :data:`vector` with :data:`index`."""
    return vector.vals[index.value]


@_make_function("remove")
def _remove(_env: ev.Environment, index: am.Numeric, vector: am.Vector) -> am.Vector:
    """Removes an item in :data:`vector` using :data:`index`."""
    vals = list(vector.vals)
    del vals[index.value]
    return am.Vector(*vals)


@_make_function("len")
def _len(_env: ev.Environment, vector: am.Vector) -> am.Numeric:
    """Returns the length of a :data:`vector`."""
    return am.Numeric(len(vector.vals))


@_make_function("cons")
def _cons(_env: ev.Environment, amalgam: am.Amalgam, vector: am.Vector) -> am.Vector:
    """Preprends an :data:`amalgam` to :data:`vector`."""
    return am.Vector(amalgam, *vector.vals)


@_make_function("snoc")
def _snoc(_env: ev.Environment, vector: am.Vector, amalgam: am.Amalgam) -> am.Vector:
    """Appends an :data:`amalgam` to :data:`vector`."""
    return am.Vector(*vector.vals, amalgam)


@_make_function("is-map")
def _is_map(_env: ev.Environment, vector: am.Vector) -> am.Atom:
    """Verifies whether :data:`vector` is a mapping."""
    if vector.mapping:
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("map-in")
def _map_in(_env: ev.Environment, vector: am.Vector, atom: am.Atom) -> am.Atom:
    """Checks whether :data:`atom` is a member of :data:`vector`."""
    if not vector.mapping:
        raise ValueError("the given vector is not a mapping")
    if atom.value in vector.mapping:
        return am.Atom("TRUE")
    return am.Atom("FALSE")


@_make_function("map-at")
def _map_at(_env: ev.Environment, vector: am.Vector, atom: am.Atom) -> am.Amalgam:
    """Obtains the value bound to :data:`atom` in :data:`vector`."""
    if not vector.mapping:
        raise ValueError("the given vector is not a mapping")
    return vector.mapping[atom.value]


@_make_function("map-up")
def _map_up(
    _env: ev.Environment,
    vector: am.Vector,
    atom: am.Atom,
    amalgam: am.Amalgam,
) -> am.Vector:
    """
    Updates the :data:`vector mapping with :data:`atom`, and
    :data:`amalgam`.
    """
    if not vector.mapping:
        raise ValueError("the given vector is not a mapping")

    new_vector: am.Vector[am.Amalgam] = am.Vector()

    mapping = {**vector.mapping}
    mapping[atom.value] = amalgam

    vals: List[am.Amalgam] = []
    for name, value in mapping.items():
        vals += (am.Atom(name), value)

    new_vector.vals = tuple(vals)
    new_vector.mapping = mapping

    return new_vector


class _Return(NamedTuple):
    """Internal utility class for signalling a return value."""
    return_value: am.Amalgam


@_make_function("return", contextual=True)
def _return(env: ev.Environment, result: am.Amalgam) -> am.Internal:
    """Exits a context with a :data:`result`."""
    return am.Internal(_Return(result))


@_make_function("break", contextual=True)
def _break(env: ev.Environment) -> am.Internal:
    """Exits a loop with :data:`:NIL`."""
    return am.Internal(_Return(am.Atom("NIL")))


@_make_function("loop", defer=True, allows=("break", "return"))
def _loop(env: ev.Environment, *qexprs: am.Quoted[am.Amalgam]) -> am.Amalgam:
    """
    Loops through and evaluates :data:`qexprs` indefinitely until a
    :data:`break` or :data:`return` is encountered.
    """
    return_value = None

    while return_value is None:
        for qexpr in qexprs:
            result = qexpr.value.evaluate(env)
            if isinstance(result, am.Internal):
                if isinstance(result.value, _Return):  # pragma: no branch
                    return_value = result.value.return_value
                break

    return return_value


@_make_function("when", defer=True)
def _when(
    env: ev.Environment, qcond: am.Quoted[am.Amalgam], qbody: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Synonym for :func:`._if` that defaults :data:`qelse` to
    :data:`:NIL`.
    """
    cond = _bool(env, qcond.value.evaluate(env))
    if cond == am.Atom("TRUE"):
        return qbody.value.evaluate(env)
    return am.Atom("NIL")


@_make_function("eval")
def _eval(env: ev.Environment, amalgam: am.Amalgam) -> am.Amalgam:
    """Evaluates a given :data:`amalgam`."""
    if isinstance(amalgam, am.Quoted):
        amalgam = amalgam.value
    return amalgam.evaluate(env)


@_make_function("unquote")
def _unquote(_env: ev.Environment, qamalgam: am.Quoted[am.Amalgam]) -> am.Amalgam:
    """Unquotes a given :data:`qamalgam`."""
    if not isinstance(qamalgam, am.Quoted):
        raise TypeError("unquotable value provided")
    return qamalgam.value


@_make_function("setr", defer=True)
def _setr(
    env: ev.Environment,
    qrname: am.Quoted[am.Symbol],
    qamalgam: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """
    Attemps to resolve :data:`qrname` to a :class:`.amalgams.Symbol`
    and binds it to the evaluated :data:`qamalgam` in the immediate
    :data:`env`.
    """
    rname = qrname.value.evaluate(env)

    if not isinstance(rname, am.Symbol):
        raise TypeError("could not resolve to a symbol")

    amalgam = qamalgam.value.evaluate(env)
    env[rname.value] = amalgam

    return amalgam


@_make_function("macro", defer=True)
def _macro(
    env: ev.Environment,
    name: am.Quoted[am.Symbol],
    args: am.Quoted[am.Vector[am.Symbol]],
    body: am.Quoted[am.Amalgam],
) -> am.Amalgam:
    """Creates a named macro using the provided arguments."""
    fn = am.create_fn(
        name.value.value,
        [arg.value for arg in args.value.vals],
        body.value,
        defer=True,
    )

    if env.parent is not None:
        fn.bind(env)

    return _setn(env, name, am.Quoted(fn))



from fractions import Fraction
import importlib.resources as resources
from io import StringIO
import re
from typing import cast, Optional

from lark import v_args, Lark, Token, Transformer, UnexpectedInput

import amalgam.amalgams as am


GRAMMAR = resources.read_text(__package__, "grammar.lark")


@v_args(inline=True)
class Expression(Transformer):
    """
    Transforms expressions in text into their respective
    :class:`.amalgams.Amalgam` representations.
    """

    def symbol(self, identifier: Token) -> am.Symbol:
        return am.Symbol(str(identifier))

    def atom(self, identifier: Token) -> am.Atom:
        return am.Atom(str(identifier))

    def integral(self, number: Token) -> am.Numeric:
        return am.Numeric(int(number))

    def floating(self, number: Token) -> am.Numeric:
        return am.Numeric(float(number))

    def fraction(self, number: Token) -> am.Numeric:
        return am.Numeric(Fraction(number))

    def string(self, *values: Token) -> am.String:
        value = "".join(values)
        value = re.sub(r"(?<!\\)\\([^\"\\])", r"\g<1>", value)
        return am.String(value.strip("\""))

    def s_expression(self, *expressions: am.Amalgam) -> am.SExpression:
        return am.SExpression(*expressions)

    def vector(self, *expressions: am.Amalgam) -> am.Vector:
        return am.Vector(*expressions)

    def quoted(self, expression: am.Amalgam) -> am.Quoted:
        return am.Quoted(expression)


class ParsingError(Exception):
    """
    Base exception for errors during parsing.

    Attributes:
      line (:class:`int`): the line number nearest to the error
      column (:class:`int`): the column number nearest to the error
    """

    def __init__(self, line: int, column: int):
        self.line = line
        self.column = column

    def __str__(self):  # pragma: no cover
        return f"near line {self.line}, column {self.column}"


class ExpectedEOF(ParsingError):
    """Raised when multiple expressions are found."""


class ExpectedExpression(ParsingError):
    """Raised when no expressions are found."""


class MissingClosing(ParsingError):
    """Raised on missing closing parentheses or brackets."""


class MissingOpening(ParsingError):
    """Raised on missing opening parentheses or brackets."""


ERROR_EXAMPLES = {
    ExpectedEOF: ("foo bar",),
    ExpectedExpression: ("",),
    MissingClosing: (
        "(", "[", "(foo", "[foo", "(foo bar", "[foo bar", "\"foo", "\"foo bar",
    ),
    MissingOpening: (")", "]", "[foo bar)", "(foo bar]"),
}


EXPR_PARSER = Lark(GRAMMAR, parser="lalr", transformer=Expression())


class Parser:
    """
    Class that serves as the frontend for parsing text.

    Attributes:
      parse_buffer (:class:`StringIO`): The text buffer used within
        :meth:`.Parser.repl_parse`.
    """

    def __init__(self) -> None:
        self.parse_buffer = StringIO()

    def repl_parse(self, text: str) -> Optional[am.Amalgam]:
        """
        Facilitates multi-line parsing for the REPL.

        Writes the given `text` string to the :attr:`parse_buffer` and
        attempts to parse `text`.

        If :class:`MissingClosing` is raised, returns `None` to allow
        for parsing to continue.

        If another subclass of :class:`ParsingError` is raised, clears
        the :attr:`parse_buffer` and re-raises the exception.

        Otherwise, if parsing succeeds, clears the :attr:`parse_buffer`
        and returns the parsed expression.
        """
        self.parse_buffer.write(text)
        self.parse_buffer.seek(0)
        text = self.parse_buffer.read()

        try:
            expr = self.parse(text)

        except MissingClosing:
            return None

        except (UnexpectedInput, ParsingError):
            self.parse_buffer = StringIO()
            raise

        else:
            self.parse_buffer = StringIO()
            return expr

    def parse(self, text: str) -> am.Amalgam:
        """Facilitates regular parsing that can fail."""
        try:
            return cast(am.Amalgam, EXPR_PARSER.parse(text))
        except UnexpectedInput as u:
            exc_cls = u.match_examples(
                EXPR_PARSER.parse, ERROR_EXAMPLES.items(),
            )
            if exc_cls is None:
                raise
            raise exc_cls(u.line, u.column) from None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['amalgam']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'lark-parser>=0.10.1,<0.11.0',
 'prompt-toolkit>=3.0.7,<4.0.0']

entry_points = \
{'console_scripts': ['amalgam = amalgam.cli:amalgam_main']}

setup_kwargs = {
    'name': 'amalgam-lisp',
    'version': '0.1.0',
    'description': 'Lisp-like interpreted language implemented in Python',
    'long_description': '<p align="center">\n  <img src="./docs/logo.png"></img>\n</p>\n\n<p align="center">\n  LISP-like interpreted language implemented in Python.\n</p>\n\n<p align="center">\n  <img src="https://img.shields.io/travis/com/purefunctor/amalgam-lisp?label=build&logo=travis&style=flat-square" href="https://travis-ci.com/github/PureFunctor/amalgam-lisp"></img>\n  <img src="https://img.shields.io/codecov/c/gh/purefunctor/amalgam-lisp?label=codecov&logo=codecov&style=flat-square" href="https://codecov.io/gh/PureFunctor/amalgam-lisp/">\n</p>\n\n# Development Setup\nInstall the following dependencies:\n* Python 3.7 & 3.8\n* [Poetry](https://python-poetry.org)\n\nClone and then navigate to the repository:\n```bash\n$ git clone https://github.com/PureFunctor/amalgam-lisp.git\n$ cd amalgam-lisp\n```\n\nInstall the dependencies for the project:\n```bash\n$ poetry install\n$ poetry run pre-commit install\n```\n\n## Running Tests / Coverage Reports / Building Documentation\n`nox` is used for the automation of the execution of tests, which generates, combines, and reports coverage data for Python 3.7 and 3.8, as well as building documentation for the project.\n```bash\n$ poetry run nox\n```\n\nAlternatively, tests, coverage reports, and the documentation can be generated manually.\n```bash\n$ poetry run coverage run -m pytest\n$ poetry run coverage combine\n$ poetry run coverage report -m\n$ poetry run coverage html\n$ poetry run sphinx-build docs docs/build\n```\n',
    'author': 'PureFunctor',
    'author_email': 'purefunctor@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PureFunctor/amalgam-lisp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from prompt_toolkit import PromptSession

import amalgam.amalgams as am
import amalgam.environment as ev
import amalgam.primordials as pd
import amalgam.parser as pr


class Engine:
    """
    Class that serves as the frontend for parsing and running programs.

    Attributes:
      parser (:class:`.parser.Parser`): A :class:`.parser.Parser`
        instance.

      environment (:class:`.environment.Environment`): An
        :class:`.environment.Environment` instance containing the
        built-in functions and a reference to the
        :class:`.engine.Engine` instance wrapped within a
        :class:`.amalgams.Internal`, accessible through the
        `~engine~` key.
    """

    def __init__(self) -> None:
        self.parser = pr.Parser()
        self.environment = ev.Environment(
            {**pd.FUNCTIONS, "~engine~": am.Internal(self)}
        )

    def repl(self, *, prompt: str = "> ", prompt_cont: str = "| ") -> None:
        """
        Runs a REPL session that supports multi-line input.

        Parameters:
          prompt (:class:`str`): The style of the prompt on
            regular lines.

          prompt_cont (:class:`str`): The style of the prompt on
            continued lines.
        """
        cont = False
        session: PromptSession = PromptSession()

        while True:
            try:
                text = session.prompt(prompt if not cont else prompt_cont)

                if cont:
                    text = "\n" + text

                expr = self.parser.repl_parse(text)

                if expr is not None:
                    print(expr.evaluate(self.environment))
                    cont = False
                else:
                    cont = True

            except EOFError:
                pd._exit(self.environment)

            except Exception as e:
                if cont:
                    cont = False
                print(f"{e.__class__.__qualname__}: {e}")

    def parse_and_run(self, text: str) -> am.Amalgam:
        """Parses and runs the given `text` string."""
        return self.parser.parse(text).evaluate(self.environment)
